/* Copyright (C) 2007 Christian Dywan <christian at twotoasts dot de>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. */

/**
 * This is an implementation of XBEL based on Glib and libXML2.
 *
 * Design Goals:
 *  - XBELitem is the only opaque public data structure.
 *  - The interface should be intuitive and familiar to Glib users.
 *  - There should be no public exposure of libXML2 specific code.
 *  - Bookmarks should actually be easily exchangeable between programs.
 *
 * TODO:
 *  - Support desc, info > metadata, alias
 *  - Compatibility: The Nokia 770 *requires* metadata and folder
 *  - Compatibility: Kazehakase's bookmarks
 *  - Compatibility: Epiphany's bookmarks
 *  - Modifying of items
 *  - Find out wether GPtrArray is a good choice
 *  - Some docs and an introduction
 *  - Optional xbel_import_file
 *  - Optional xbel_export_file
 *  - Consider a more liberate license
 **/

#include "xbel.h"

// Private: Trim leading and trailing whitespace in place.
static gchar* xbel_trim(gchar* sString)
{
    // Based on trim.c by Bob Stout
    g_return_val_if_fail(sString != NULL, NULL);
    gchar* sBufferA = sString; gchar* sBufferB = sString;
    for(sBufferA = sString; *sBufferA && isspace(*sBufferA); ++sBufferA)
     ;
    if(sString != sBufferA)
     memmove(sString, sBufferA, sBufferA - sString);
    guint i = 0; guint iCount = 0;
    while(*sBufferA)
    {
     if(isspace(*sBufferA) && iCount)
      sBufferA++;
     else
     {
      if(!isspace(*sBufferA))
       iCount = 0;
      else
      {
       *sBufferA = ' ';
       iCount = 1;
      }
      sBufferB[i++] = *sBufferA++;
     }
    }
    sBufferB[i] = '\0';
    while(--i >= 0)
     if(!isspace(sBufferB[i]))
      break;
    sBufferB[++i] = '\0';
    return sString;
}

// Private: Create a new item of a certain type
static XBELitem* xbel_item_new(XBELitemType type)
{
    XBELitem* item = g_new(XBELitem, 1);
    item->parent = NULL;
    item->type = type;
    if(type == XBEL_ITEM_FOLDER) item->items = g_ptr_array_new();
    if(type != XBEL_ITEM_SEPARATOR)
    {
     item->sTitle = NULL;
     item->sDesc  = NULL;
    }
    if(type == XBEL_ITEM_BOOKMARK) item->sHref = NULL;
    return item;
}

/**
 * Create a new empty bookmark.
 *
 * @return: a newly allocated bookmark
 **/
XBELitem* xbel_bookmark_new(void)
{
    return xbel_item_new(XBEL_ITEM_BOOKMARK);
}

static XBELitem* xbel_bookmark_from_xmlNodePtr(xmlNodePtr cur)
{
    g_return_if_fail(cur != NULL);
    XBELitem* bookmark = xbel_bookmark_new();
    xmlChar* key = xmlGetProp(cur, "href");
    bookmark->sHref = xbel_trim(key);
    cur = cur->xmlChildrenNode;
    while(cur != NULL)
    {
     if(!xmlStrcmp(cur->name, (const xmlChar*)"title"))
     {
      xmlChar* key = xmlNodeGetContent(cur);
      bookmark->sTitle = g_strdup(xbel_trim(key));
      xmlFree(key);
     }
     cur = cur->next;
    }
    return bookmark;
}

/**
 * Create a new separator.
 *
 * The returned item must be freed eventually.
 *
 * @return: a newly allocated separator.
 **/
XBELitem* xbel_separator_new(void)
{
    return xbel_item_new(XBEL_ITEM_SEPARATOR);
}

/**
 * Create a new empty folder.
 *
 * The returned item must be freed eventually.
 *
 * @return: a newly allocated folder.
 **/
XBELitem* xbel_folder_new(void)
{
    return xbel_item_new(XBEL_ITEM_FOLDER);
}

// Private: Create a folder from an xmlNodePtr
static XBELitem* xbel_folder_from_xmlNodePtr(xmlNodePtr cur)
{
    g_return_if_fail(cur != NULL);
    XBELitem* folder = xbel_folder_new();
    cur = cur->xmlChildrenNode;
    while(cur != NULL)
    {
     if(!xmlStrcmp(cur->name, (const xmlChar*)"title"))
     {
      xmlChar* key = xmlNodeGetContent(cur);
      folder->sTitle = xbel_trim(key);
     }
     else if(!xmlStrcmp(cur->name, (const xmlChar*)"bookmark"))
     {
      XBELitem* item = xbel_bookmark_from_xmlNodePtr(cur);
      item->parent = (struct XBELitem*)folder;
      g_ptr_array_add(folder->items, item);
     }
     cur = cur->next;
    }
    return folder;
}

// Private: Loads the contents from an xmlNodePtr into a folder.
static gboolean xbel_folder_from_xmlDocPtr(XBELitem* folder, xmlDocPtr doc)
{
    g_return_val_if_fail(doc != NULL, FALSE);
    g_return_val_if_fail(xbel_item_is_folder(folder), FALSE);
    if(folder->items->len)
     g_warning("The folder is not empty.");
    xmlNodePtr cur = xmlDocGetRootElement(doc);
    if((cur = xmlDocGetRootElement(doc)) == NULL)
    {
     // Empty document
     return FALSE;
    }
    if(xmlStrcmp(cur->name, (const xmlChar*)"xbel"))
    {
     // Wrong document type
     return FALSE;
    }
    cur = cur->xmlChildrenNode;
    while(cur != NULL)
    {
     XBELitem* item = NULL;
     if(!xmlStrcmp(cur->name, (const xmlChar*)"folder"))
      item = xbel_folder_from_xmlNodePtr(cur);
     else if(!xmlStrcmp(cur->name, (const xmlChar*)"bookmark"))
      item = xbel_bookmark_from_xmlNodePtr(cur);
     else if(!xmlStrcmp(cur->name, (const xmlChar*)"separator"))
      item = xbel_separator_new();
     /*else if(!xmlStrcmp(cur->name, (const xmlChar*)"info"))
      item = xbel_parse_info(xbel, cur);*/
     if(item != NULL)
     {
      item->parent = (struct XBELitem*)folder;
      g_ptr_array_add(folder->items, item);
     }
     cur = cur->next;
    }
    return TRUE;
}

/**
 * Free an XBELitem you own.
 *
 * Freeing items contained in a folder is not allowed.
 **/
void xbel_item_free(XBELitem* item)
{
    g_return_if_fail(item != NULL);
    if(item->parent != NULL)
    {
     g_warning("Mustn't free an item contained in a folder.");
     return;
    }
    if(xbel_item_is_folder(item))
    {
     guint i; for(i = 0; i < item->items->len; i++)
     {
      XBELitem* _item = g_ptr_array_index(item->items, i);
      _item->parent = NULL;
      xbel_item_free(_item);
     }
     g_ptr_array_free(item->items, TRUE);
    }
    // FIXME: Free other values
}

/**
 * Copy an XBELitem.
 *
 * The returned item must be freed eventually.
 *
 * @item: the item to copy
 * @return: a copy of item
 **/
XBELitem* xbel_item_copy(XBELitem* item)
{
    // FIXME: Return a copy of the item
    g_warning("Copying is not currently implemented.");
    return NULL;
}

/**
 * Append the given item to a folder.
 *
 * The item is actually moved and must not be contained in another folder.
 *
 * @folder: a folder
 * @item:   the item to append
 **/
void xbel_folder_append_item(XBELitem* folder, XBELitem* item)
{
    g_return_if_fail(xbel_item_get_parent(item) == NULL);
    g_return_if_fail(xbel_item_is_folder(folder));
    // FIXME: Should undergo a few tests
    item->parent = (struct XBELitem*)folder;
    g_ptr_array_add(folder->items, item);
}

/**
 * Remove the given item from a folder.
 *
 * @folder: a folder
 * @item:   the item to remove
 * @return: the item on success or NULL if the item wasn't found
 **/
XBELitem* xbel_folder_remove_item(XBELitem* folder, XBELitem* item)
{
    g_return_if_fail(item != NULL);
    g_return_if_fail(xbel_item_get_parent(folder) != item);
    // FIXME: Should undergo a few tests
    item->parent = NULL;
    // Fortunately we know that items are unique
    return g_ptr_array_remove(folder->items, item) ? item : NULL;
}

/**
 * Retrieve the number of items contained in a folder.
 *
 * @folder: a folder
 * @return: number of items
 **/
guint xbel_folder_get_n_items(XBELitem* folder)
{
    g_return_val_if_fail(xbel_item_is_folder(folder), 0);
    return folder->items->len;
}

/**
 * Retrieve an item contained in a folder by its index.
 *
 * @folder: a folder
 * @return: the item at the given index
 **/
XBELitem* xbel_folder_get_nth_item(XBELitem* folder, guint n)
{
    g_return_val_if_fail(xbel_item_is_folder(folder), NULL);
    return g_ptr_array_index(folder->items, n);
}

/**
 * @item:   A valid item.
 * @return: The type of the item.
 **/
XBELitemType xbel_item_get_type(XBELitem* item)
{
    return item->type;
}

/**
 * @item:   A valid item.
 * @return: The parent folder of the item.
 **/
XBELitem* xbel_item_get_parent(XBELitem* item)
{
    g_return_val_if_fail(item != NULL, NULL);
    return (XBELitem*)item->parent;
}

G_CONST_RETURN gchar* xbel_item_get_title(XBELitem* item)
{
    g_return_val_if_fail(!xbel_item_is_separator(item), NULL);
    return item->sTitle;
}

G_CONST_RETURN gchar* xbel_item_get_desc(XBELitem* item)
{
    g_return_val_if_fail(!xbel_item_is_separator(item), NULL);
    return item->sDesc;
}

G_CONST_RETURN gchar* xbel_bookmark_get_href(XBELitem* item)
{
    g_return_val_if_fail(xbel_item_is_bookmark(item), NULL);
    return item->sHref;
}

gboolean xbel_item_is_bookmark(XBELitem* item)
{
    g_return_val_if_fail(item != NULL, FALSE);
    return item->type == XBEL_ITEM_BOOKMARK;
}

gboolean xbel_item_is_separator(XBELitem* item)
{
    g_return_val_if_fail(item != NULL, FALSE);
    return item->type == XBEL_ITEM_SEPARATOR;
}

gboolean xbel_item_is_folder(XBELitem* item)
{
    g_return_val_if_fail(item != NULL, FALSE);
    return item->type == XBEL_ITEM_FOLDER;
}

gboolean xbel_folder_from_data(XBELitem* folder, const gchar* sData
 , GError** error)
{
    g_return_val_if_fail(sData != NULL, FALSE);
    g_return_val_if_fail(xbel_item_is_folder(folder), FALSE);
    if(folder->items->len)
     g_warning("The folder is not empty.");
    xmlDocPtr doc;
    if((doc = xmlParseMemory(sData, strlen(sData))) == NULL)
    {
     // No valid xml or broken encoding
     *error = g_error_new(XBEL_ERROR, XBEL_ERROR_READ
      , "Malformed document.");
     return FALSE;
    }
    if(!xbel_folder_from_xmlDocPtr(folder, doc))
    {
     // Parsing failed
     xmlFreeDoc(doc);
     *error = g_error_new(XBEL_ERROR, XBEL_ERROR_READ
      , "Malformed document.");
     return FALSE;
    }
    xmlFreeDoc(doc);
    return TRUE;
}

gboolean xbel_folder_from_file(XBELitem* folder, const gchar* sFilename
 , GError** error)
{
    g_return_val_if_fail(sFilename != NULL, FALSE);
    g_return_val_if_fail(xbel_item_is_folder(folder), FALSE);
    if(folder->items->len)
     g_warning("The folder is not empty.");
    xmlDocPtr doc;
    if((doc = xmlParseFile(sFilename)) == NULL)
    {
     // No valid xml or broken encoding
     *error = g_error_new(XBEL_ERROR, XBEL_ERROR_READ
      , "Malformed document.");
     return FALSE;
    }
    if(!xbel_folder_from_xmlDocPtr(folder, doc))
    {
     // Parsing failed
     xmlFreeDoc(doc);
     *error = g_error_new(XBEL_ERROR, XBEL_ERROR_READ
      , "Malformed document.");
     return FALSE;
    }
    xmlFreeDoc(doc);
    return TRUE;
}

gboolean xbel_folder_from_data_dirs(XBELitem* folder, const gchar* sFilename
 , gchar** sPath, GError** error)
{
    g_return_val_if_fail(sFilename != NULL, FALSE);
    g_return_val_if_fail(xbel_item_is_folder(folder), FALSE);
    if(folder->items->len)
     g_warning("The folder is not empty.");
    // FIXME: Essentially unimplemented

    *error = g_error_new(XBEL_ERROR, XBEL_ERROR_READ
     , "Malformed document.");
    return FALSE;
}

gchar* xbel_folder_to_data(XBELitem* folder, gsize* length, GError** error)
{
    g_return_val_if_fail(xbel_item_is_folder(folder), FALSE);
    // FIXME: Essentially unimplemented

    *error = g_error_new(XBEL_ERROR, XBEL_ERROR_WRITE
     , "Not currently implemented."/*"Writing failed."*/);
    return NULL;
}

gboolean xbel_to_file(XBELitem* folder, const gchar* sFilename, GError** error)
{
    gchar* sData;
    if((sData = xbel_folder_to_data(folder, NULL, error)) == NULL)
     return FALSE;
    FILE* fp;
    // FIXME: Should undergo a few tests
    if((fp = fopen(sFilename, "w")) == NULL)
    {
     *error = g_error_new(G_FILE_ERROR, G_FILE_ERROR_ACCES
     , "Writing failed.");
     return FALSE;
    }
    fputs(sData, fp);
    fclose(fp);
    g_free(sData);
    /*xmlSaveFormatFile(sFilename, doc, 1);*/
    return TRUE;
}
